from __future__ import annotations

import concurrent.futures
import glob
import os
import os.path as op
import re
import shutil
from pathlib import Path
from typing import Callable
from typing import cast
from typing import Sequence
from urllib.parse import unquote_plus
from urllib.parse import urlparse
from urllib.parse import urlunparse

from google.cloud.storage import Blob
from google.cloud.storage import Bucket
from google.cloud.storage import Client


__version__ = "1.2.0"


def copy_file(
    src: str | Path | Blob,
    dst: str | Path | Blob,
    *,
    client: Client | None = None,
    quiet: bool = False,
) -> None:
    """Copy a single file.

    If `src` and `dst` are both determined to be local files then `client` is ignored.
    """
    return _copy(src, dst, _scheme_copy_fns=_FILE_FUNCTIONS, client=client, quiet=quiet)


def copy_dir(
    src: str | Path | Blob,
    dst: str | Path | Blob,
    *,
    client: Client | None = None,
    quiet: bool = False,
) -> None:
    """Copy a directory (recursively).

    If `src` and `dst` are both determined to be local directories
    then `client` is ignored.
    """
    return _copy(src, dst, _scheme_copy_fns=_DIR_FUNCTIONS, client=client, quiet=quiet)


def copy_files(
    srcs: Sequence[str | Path | Blob],
    dsts: str | Path | Blob | Sequence[str | Path | Blob],
    *,
    client: Client | None = None,
    quiet: bool = False,
) -> None:
    """Copy a list of files.

    If `dsts` is a `str | Path | Blob` it is treated as a directory
    and each of the files in `srcs` will have its name "flattened" and will be
    copied under `dsts`.

    if `dsts` is a `Sequence[str | Path | Blob]` it is zipped with `srcs`, i.e.
    each file in `srcs` is copied to its corresponding entry in `dsts`.
    """
    if isinstance(dsts, (str, Path)):
        _dsts = [op.join(dsts, _flatten(src)) for src in srcs]
    elif isinstance(dsts, Blob):
        blob_name: str = dsts.name or ""  # type: ignore
        bucket: Bucket = dsts.bucket  # type: ignore
        names = [op.join(blob_name, _flatten(src)) for src in srcs]
        _dsts = [Blob(name, bucket) for name in names]
    else:
        _dsts = dsts

    with concurrent.futures.ThreadPoolExecutor() as executor:
        # submit each of the file-copy jobs to the thread pool
        future_to_src: dict[concurrent.futures.Future[None], str | Path | Blob] = {}
        for src, dst in zip(srcs, _dsts):
            future = executor.submit(copy_file, src, dst, client=client, quiet=True)
            future_to_src[future] = src

        # report the status of each copied file
        completed_futures = concurrent.futures.as_completed(future_to_src)
        for i, future in enumerate(completed_futures, 1):
            src = future_to_src[future]
            future.result()
            if not quiet:
                _log_successful_copy(src, n=i, N=len(future_to_src))


# --- PRIVATE API ---


def _copy(
    src: str | Path | Blob,
    dst: str | Path | Blob,
    *,
    _scheme_copy_fns: dict[tuple[str, str], Callable],
    client: Client | None = None,
    quiet: bool = False,
) -> None:
    src_scheme = _parse_scheme(src)
    dst_scheme = _parse_scheme(dst)
    schemes = (src_scheme, dst_scheme)

    copy_fn = _scheme_copy_fns[schemes]

    # local copy operation
    if "gs" not in schemes:
        copy_fn(src, dst, quiet=quiet)  # type: ignore
        return

    # one (or both) of src/dst are remote URIs, need to
    # forward the storag client
    client = client or Client()
    copy_fn(src, dst, client=client, quiet=quiet)  # type: ignore


def _parse_scheme(arg: str | Path | Blob) -> str:
    if isinstance(arg, str):
        scheme = urlparse(arg).scheme
        if scheme in ("", "file"):
            return ""
        elif scheme == "gs":
            return "gs"
        else:
            raise ValueError(f"Failed to determine scheme for {arg!r}")
    elif isinstance(arg, Path):
        return ""
    elif isinstance(arg, Blob):
        return "gs"
    else:
        raise ValueError(f"Failed to determine scheme for {arg!r}")


# --- UTILITY FUNCTIONS ---


def _blob_to_uri(blob: Blob) -> str:
    components = ("gs", blob.bucket.name, blob.name, "", "", "")
    return urlunparse(components)


def _filename_to_uri(filename: str | Path) -> str:
    components = ("file", "", str(filename), "", "", "")
    return urlunparse(components)


def _uri_to_filename(uri: str | Path) -> str:
    uri = unquote_plus(urlparse(uri).path) if isinstance(uri, str) else uri
    return str(uri)


def _log_successful_copy(
    src: str | Path | Blob,
    *,
    n: int | None = 1,
    N: int | None = 1,
):
    uri = _filename_to_uri(src) if isinstance(src, (str, Path)) else _blob_to_uri(src)
    prefix = "" if None in (n, N) else f"[{n}/{N}] - "
    print(f"{prefix}Copied {uri!r}")


def _log_skipping_file(
    src: str | Path,
    *,
    n: int | None = 1,
    N: int | None = 1,
) -> None:
    uri = _filename_to_uri(src)
    prefix = "" if None in (n, N) else f"[{n}/{N}] - "
    print(f"{prefix}Skipping {uri!r}")


def _flatten(spb: str | Path | Blob) -> str:
    """Convert a URI (local or remote) to a flat filename.

    Examples:
        >>> _flatten('.cache/dir/file.txt')
        '.cache-dir-file.txt'
        >>> _flatten('/abs/path to  some/file.tar.gz')
        'abs-path-to-some-file.tar.gz'
        >>> _flatten('gs://bkt/some/blob/')
        'gs-bkt-some-blob'
        >>> _flatten('gs://bkt')
        'gs-bkt'
        >>> # can pass Path objects:
        >>> from pathlib import Path
        >>> fp = Path('path/to/my/file.csv')
        >>> _flatten(fp)
        'path-to-my-file.csv'
        >>> # can pass storage.Blob
        >>> from google.cloud.storage import Blob
        >>> blob = Blob.from_string('gs://bkt/my/module.py')
        >>> _flatten(blob)
        'bkt-my-module.py'
    """
    s = str(spb) if isinstance(spb, (str, Path)) else _blob_to_uri(spb)
    filename_s = re.sub(r"[^0-9a-zA-Z_\.-]+", "-", s)  # remove non-(alnum | _ | . | -)
    filename_s = re.sub(r"(^-+|-+$)", "", filename_s)  # remove leading/trailing '-'
    return filename_s


# --- COPY FUNCTION IMPLEMENTATIONS ---


def _copy_file(src: str | Path, dst: str | Path, *, quiet: bool = False):
    """local file -> local file"""
    shutil.copy2(src, dst)
    if not quiet:
        _log_successful_copy(src, n=None, N=None)


def _sync_files(src: str | Path, dst: str | Path, *, quiet: bool = False):
    """local dir -> local dir"""

    # NOTE: to support python3.7 we do our own recursive copying
    # If support for python3.7 is dropped, change this back to the
    # implementation here:
    # https://github.com/andrewrosss/gcs-uri/blob/0fd3d4b8fbae947721c9e099a1239e3c551d13ed/src/gcs_uri.py#L213-L222  # noqa: E501
    srcroot = Path(src)
    srcpaths = list(srcroot.rglob("*"))
    for i, srcpath in enumerate(srcpaths):
        relpath = srcpath.relative_to(srcroot)
        dstpath = Path(dst) / relpath
        if srcpath.is_dir():
            dstpath.mkdir(exist_ok=True)
        elif srcpath.is_file():
            shutil.copy2(srcpath, dstpath)
            if not quiet:
                _log_successful_copy(srcpath, n=i, N=len(srcpaths))
        else:
            if not quiet:
                _log_skipping_file(srcpath, n=i, N=len(srcpaths))


def _download_file(
    src: str | Blob,
    dst: str | Path,
    *,
    client: Client | None = None,
    quiet: bool = False,
):
    """remote blob -> local file"""
    client = client or Client()
    _src = Blob.from_string(src, client=client) if isinstance(src, str) else src
    _dst = _uri_to_filename(dst)
    if op.isdir(_dst):
        filename = op.join(_dst, op.basename(_src.name))
    else:
        filename = _dst
    _src.download_to_filename(filename, client=client)
    if not quiet:
        _log_successful_copy(_src)


def _upload_file(
    src: str | Path,
    dst: str | Blob,
    *,
    client: Client | None = None,
    quiet: bool = False,
):
    """local file -> remote blob"""
    client = client or Client()
    _src = _uri_to_filename(src)
    _dst = Blob.from_string(dst, client=client) if isinstance(dst, str) else dst
    if cast(str, _dst.name).endswith("/"):
        _dst.name = op.join(_dst.name, op.basename(_src))
    _dst.upload_from_filename(_src, client=client)
    if not quiet:
        _log_successful_copy(_src)


def _download_dir(
    src: str | Blob,
    dst: str | Path,
    *,
    client: Client | None = None,
    quiet: bool = False,
):
    """remote dir -> local dir"""
    _client = client or Client()
    _src = Blob.from_string(src, client=_client) if isinstance(src, str) else src
    _dst = _uri_to_filename(dst)
    blobs: list[Blob] = list(_client.list_blobs(_src.bucket, prefix=_src.name))

    with concurrent.futures.ThreadPoolExecutor() as executor:
        # submit each of the download jobs to the thread pool
        future_to_uri = {}
        for b in blobs:
            relpath = op.relpath(b.name, _src.name or "")
            filename = op.join(_dst, relpath)  # type: ignore
            os.makedirs(op.dirname(filename), exist_ok=True)
            future = executor.submit(b.download_to_filename, filename, client=client)
            future_to_uri[future] = _blob_to_uri(b)

        # report the status of each downloaded file
        completed_futures = concurrent.futures.as_completed(future_to_uri)
        for i, future in enumerate(completed_futures, 1):
            uri = future_to_uri[future]
            future.result()
            if not quiet:
                _log_successful_copy(uri, n=i, N=len(future_to_uri))


def _upload_dir(
    src: str | Path,
    dst: str | Blob,
    *,
    client: Client | None = None,
    quiet: bool = False,
):
    """local dir -> remote dir"""
    client = client or Client()
    _src = _uri_to_filename(src)
    _dst = Blob.from_string(dst, client=client) if isinstance(dst, str) else dst
    pattern = op.join(_src, "**")
    files = [f for f in glob.glob(pattern, recursive=True) if op.isfile(f)]

    with concurrent.futures.ThreadPoolExecutor() as executor:
        # submit each of the download jobs to the thread pool
        future_to_filename = {}
        for filename in files:
            relpath = op.relpath(filename, _src)
            name = op.join(_dst.name, relpath)
            b = Blob(name, _dst.bucket)
            future = executor.submit(b.upload_from_filename, filename, client=client)
            future_to_filename[future] = filename

        # report the status of each downloaded file
        completed_futures = concurrent.futures.as_completed(future_to_filename)
        for i, future in enumerate(completed_futures, 1):
            filename = future_to_filename[future]
            future.result()
            if not quiet:
                _log_successful_copy(filename, n=i, N=len(future_to_filename))


def _copy_blob(
    src: str | Blob,
    dst: str | Blob,
    *,
    client: Client | None = None,
    quiet: bool = False,
):
    """remote blob -> remote blob"""
    client = client or Client()
    _src = Blob.from_string(src, client=client) if isinstance(src, str) else src
    _dst = Blob.from_string(dst, client=client) if isinstance(dst, str) else dst
    src_is_dir = cast(str, _src.name).endswith("/")
    dst_is_dir = cast(str, _dst.name).endswith("/")
    if dst_is_dir and not src_is_dir:
        _dst.name = op.join(_dst.name, op.basename(_src.name))
    _src.bucket.copy_blob(_src, _dst.bucket, _dst.name, client=client)
    if not quiet:
        _log_successful_copy(_dst)


def _sync_blobs(
    src: str | Blob,
    dst: str | Blob,
    *,
    client: Client | None = None,
    quiet: bool = False,
):
    """remote dir -> remote dir"""
    _client = client or Client()
    _src = Blob.from_string(src, client=_client) if isinstance(src, str) else src
    _dst = Blob.from_string(dst, client=_client) if isinstance(dst, str) else dst
    src_blobs: list[Blob] = list(_client.list_blobs(_src.bucket, prefix=_src.name))

    with concurrent.futures.ThreadPoolExecutor() as executor:
        # submit each of the download jobs to the thread pool
        future_to_blob: dict[concurrent.futures.Future[None], Blob] = {}
        for src_blob in src_blobs:
            relpath = op.relpath(src_blob.name, _src.name)
            new_name = op.join(_dst.name, relpath)
            dst_blob = Blob(new_name, _dst.bucket)
            _src_uri = _blob_to_uri(src_blob)
            _dst_uri = _blob_to_uri(dst_blob)
            future = executor.submit(
                _copy_blob,
                _src_uri,
                _dst_uri,
                client=_client,
                quiet=True,
            )
            future_to_blob[future] = src_blob

        # report the status of each downloaded file
        completed_futures = concurrent.futures.as_completed(future_to_blob)
        for i, future in enumerate(completed_futures, 1):
            blob = future_to_blob[future]
            future.result()
            if not quiet:
                _log_successful_copy(blob, n=i, N=len(future_to_blob))


# --- MAPPING OF SCHEMES TO COPY FUNCTIONS ---

_FILE_FUNCTIONS = {
    ("", ""): _copy_file,  # local file -> local file
    ("gs", ""): _download_file,  # remote file -> local file
    ("", "gs"): _upload_file,  # local file -> remote file
    ("gs", "gs"): _copy_blob,  # remote file -> remote file
}

_DIR_FUNCTIONS = {
    ("", ""): _sync_files,  # local dir -> local dir
    ("gs", ""): _download_dir,  # remote dir -> local dir
    ("", "gs"): _upload_dir,  # local dir -> remote dir
    ("gs", "gs"): _sync_blobs,  # remote dir -> remote dir
}



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

modules = \
['gcs_uri']
install_requires = \
['google-cloud-storage>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'gcs-uri',
    'version': '1.2.0',
    'description': 'Simple API to copy files to and from Google Cloud Storage',
    'long_description': '# gcs-uri\n\nSimple API to copy files to and from Google Cloud Storage\n\n[![PyPI Version](https://img.shields.io/pypi/v/gcs-uri.svg)](https://pypi.org/project/gcs-uri/)\n\n## Installation\n\n```bash\npip install gcs-uri\n```\n\n## Usage\n\n`gcs-uri` exposes the following functions as its main public API\n\n- `copy_file`\n- `copy_dir`\n- `copy_files`\n\nThese functions do exactly what they sound like they do.\n\n`copy_file` will copy a source file (either a local file or a remote blob in GCS) to destination file (either a local file or remote blob in GCS).\n\n`copy_dir` will recursively copy the contents of a directory (either a local directory or a remote "directory" in GCS) to a destination directory (either a local directory or a remote "directory" in GCS)\n\n`copy_files` will copy a list of source files (either local files or remote blobs in GCS or a mix of local files/remote blobs) to a corresponding set of destination files (either local files or remote blobs in GCS of a mix of local files/remote blobs)\n\n> If the second argument to `copy_files` is of type `str | Path | Blob` (as opposed to a Sequence), then this argument is treated like a directory and each of the source files are "flattened" (i.e. folder delimiters are removed) and copied under the destintation directory.\n\nThe idea being that you can pass just about any object to these functions and the functions will figures how to do the copying.\n\n## Examples\n\n### Local file -> local file\n\nIn this case `copy_file` behaves just like `shutil.copy2` or `cp`, copying the source file to the destination file locally.\n\n```python\nsrc = \'/my/src/file.txt\'\ndst = \'/my/dst/file.txt\'\n\ncopy_file(src, dst)\n```\n\n`src` and `dst` can also be `pathlib.Path` objects:\n\n```python\nfrom pathlib import Path\n\nsrc = Path(\'/my/src/file.txt\')\ndst = Path(\'/my/dst/file.txt\')\n\ncopy_file(src, dst)\n```\n\n### Local dir -> local dir\n\nIn this case `copy_dir` behaves just like `shutil.copytree` (or somewhat like rsync, but `copy_dir` will "re-copy" all files to the destination whether they exist in the the destination or not).\n\n```python\nsrc = \'/my/src\'\ndst = \'/my/dst\'\n\ncopy_dir(src, dst)\n\n# if there was a file `/my/src/a/b.txt` after `copy_dir`\n# there would then be a file `/my/dst/a/b.txt`\n```\n\nThe source and destination can include or omit a trailing slash and the results are the same as above.\n\n### Local file -> remote file (upload)\n\nTo copy a file to a google cloud bucket, barely anything has to change, the destination should simply be a google storage URI:\n\n```python\nsrc = \'/my/src/file.txt\'\ndst = \'gs://my-bkt/dst/file.txt\'\n\ncopy_file(src, dst)\n```\n\nIf you would like `gcs-uri` to use a particular Google Storage Client, this can be provided as a keyword(-only) argument (the same applies to `copy_dir`):\n\n```python\nfrom google.cloud import storage\n\nclient = storage.Client()\n\nsrc = \'/my/src/file.txt\'\ndst = \'gs://my-bkt/dst/file.txt\'\n\ncopy_file(src, dst, client=client)\n```\n\nIf no client is provided and either of the source or destinations (or both) are determined to represent a remote location then `gcs-uri` will try to instantiate a client by calling `storage.Client()`.\n\nNote, we can provided `gcs-uri` with "richer" objects (instead of just strings):\n\n```python\nfrom pathlib import Path\nfrom google.cloud import storage\n\nclient = storage.Client()\n\nsrc = Path(\'/my/src/file.txt\')\ndst = storage.Blob.from_string(\'gs://my-bkt/dst/file.txt\', client=client)\n\ncopy_file(src, dst)\n```\n\n### Local dir -> remote dir (upload)\n\nThe concepts from the previous sections apply here:\n\n```python\nsrc = \'/my/src\'\ndst = \'gs://my-bkt/dst\'\n\ncopy_dir(src, dst)\n\n# if there was a file `/my/src/a/b.txt` after `copy_dir`\n# there would then be a blob `gs://my-bkt/dst/a/b.txt`\n```\n\n### Remote file -> local file (download)\n\n```python\nsrc = \'gs://my-bkt/src/file.txt\'\ndst = \'/my/dst/file.txt\'\n\ncopy_file(src, dst)\n```\n\n### Remote dir -> local dir (download)\n\n```python\nsrc = \'gs://my-bkt/src\'\ndst = \'/my/dst\'\n\ncopy_dir(src, dst)\n```\n\n### Remote file -> remote file (transfer)\n\n```python\nsrc = \'gs://my-bkt/src/file.txt\'\ndst = \'gs://my-other-bkt/dst/file.txt\'\n\ncopy_file(src, dst)\n```\n\n### Remote dir -> remote dir (transfer)\n\n```python\nsrc = \'gs://my-bkt/src\'\ndst = \'gs://my-other-bkt/dst\'\n\ncopy_dir(src, dst)\n```\n\n### List of local files -> list of remote files\n\n```python\nsrcs = [\'/my/src/file1.txt\', \'/my/src/file2.txt\']\ndsts = [\'gs://my-bkt/dst/file1.txt\', \'gs://my-bkt/dst/file2.txt\']\n\ncopy_files(srcs, dsts)\n# copies: /my/src/file1.txt -> gs://my-bkt/dst/file1.txt\n# copies: /my/src/file2.txt -> gs://my-bkt/dst/file2.txt\n```\n\n### List of local files -> remote dir\n\n```python\nsrcs = [\'/my/src/file1.txt\', \'/my/src/file2.txt\']\ndst = \'gs://my-bkt/dst\'\n\ncopy_files(srcs, dst)\n# copies: /my/src/file1.txt -> gs://my-bkt/dst/my-src-file1.txt\n# copies: /my/src/file2.txt -> gs://my-bkt/dst/my-src-file2.txt\n```\n\n## API\n\n```python\n# src/gcs_uri.py\n\ndef copy_file(\n    src: str | Path | Blob,\n    dst: str | Path | Blob,\n    *,\n    client: Client | None = None,\n    quiet: bool = False,\n) -> None:\n    """Copy a single file.\n\n    If `src` and `dst` are both determined to be local files then `client` is ignored.\n    """\n\ndef copy_dir(\n    src: str | Path | Blob,\n    dst: str | Path | Blob,\n    *,\n    client: Client | None = None,\n    quiet: bool = False,\n) -> None:\n    """Copy a directory (recursively).\n\n    If `src` and `dst` are both determined to be local directories\n    then `client` is ignored.\n    """\n\ndef copy_files(\n    srcs: Sequence[str | Path | Blob],\n    dsts: str | Path | Blob | Sequence[str | Path | Blob],\n    *,\n    client: Client | None = None,\n    quiet: bool = False,\n) -> None:\n    """Copy a list of files.\n\n    If `dsts` is a `str | Path | Blob` it is treated as a directory\n    and each of the files in `srcs` will have its name "flattened" and will be\n    copied under `dsts`.\n\n    if `dsts` is a `Sequence[str | Path | Blob]` it is zipped with `srcs`, i.e.\n    each file in `srcs` is copied to its corresponding entry in `dsts`.\n    """\n```\n\n## Tests\n\nThis package comes with some basic end-to-end (e2e) tests. They require an active google cloud project with the google storage API enabled.\n\nTo help with running them there is a utility script in the root of this repo: `run_e2e_tests.py`.\n\n```text\nusage: run_e2e_tests.py [-h] [-v] [-c GOOGLE_APPLICATION_CREDENTIALS]\n                        [-u TEST_STORAGE_URI]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -v, --version         show program\'s version number and exit\n  -c GOOGLE_APPLICATION_CREDENTIALS, --google-application-credentials GOOGLE_APPLICATION_CREDENTIALS\n                        Google cloud service account to use.\n  -u TEST_STORAGE_URI, --test-storage-uri TEST_STORAGE_URI\n                        Google storage uri to use when running e2e tests.\n```\n\nThis script requires you to provided a service account json file as we\'ll as a URI to a location in google cloud which the tests will use to copy blobs to/from. (**IMPORTANT**: **_all_** blobs at and beneath the location you specifify will be removed - the bucket itself will **not** be removed).\n\nSo, run the e2e tests with something like:\n\n```bash\npython -m run_e2e_tests -c "path/to/service-account.json" -u "gs://my-bkt/gcs-uri-tests"\n```\n\n## Contributing\n\n1. Have or install a recent version of `poetry` (version >= 1.1)\n1. Fork the repo\n1. Setup a virtual environment (however you prefer)\n1. Run `poetry install`\n1. Run `pre-commit install`\n1. Add your changes (adding/updating tests is always nice too)\n1. Commit your changes + push to your fork\n1. Open a PR\n',
    'author': 'Andrew Ross',
    'author_email': 'andrew.ross.mail@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/andrewrosss/gcs-uri',
    'package_dir': package_dir,
    'py_modules': modules,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
